@@ -25,10 +25,10 @@ class Elastic(torch.nn.Module):
 
     Args:
         lamb:
-            A 2D Tensor containing an initial guess of the first Lame'
+            A 2D Tensor containing an initial guess of the first Lamé
             parameter (lambda).
         mu:
-            A 2D Tensor containing an initial guess of the second Lame'
+            A 2D Tensor containing an initial guess of the second Lamé
             parameter.
         buoyancy:
             A 2D Tensor containing an initial guess of the buoyancy
@@ -62,13 +62,18 @@ class Elastic(torch.nn.Module):
             raise RuntimeError("buoyancy must have two dimensions")
         self.lamb = torch.nn.Parameter(lamb, requires_grad=lamb_requires_grad)
         self.mu = torch.nn.Parameter(mu, requires_grad=mu_requires_grad)
-        self.buoyancy = torch.nn.Parameter(buoyancy,
-                                           requires_grad=buoyancy_requires_grad)
+        self.buoyancy = torch.nn.Parameter(
+            buoyancy, requires_grad=buoyancy_requires_grad
+        )
         self.grid_spacing = grid_spacing
 
-    def forward(self, dt: float, source_amplitudes: Optional[Tensor] = None,
-                source_locations: Optional[Tensor] = None,
-                receiver_locations: Optional[Tensor] = None,
+    def forward(self, dt: float,
+                source_amplitudes_y: Optional[Tensor] = None,
+                source_amplitudes_x: Optional[Tensor] = None,
+                source_locations_y: Optional[Tensor] = None,
+                source_locations_x: Optional[Tensor] = None,
+                receiver_locations_y: Optional[Tensor] = None,
+                receiver_locations_x: Optional[Tensor] = None,
                 accuracy: int = 4, pml_width: Union[int, List[int]] = 20,
                 pml_freq: Optional[float] = None,
                 max_vel: Optional[float] = None,
@@ -102,6 +107,7 @@ class Elastic(torch.nn.Module):
                                                                     Tensor,
                                                                     Tensor,
                                                                     Tensor,
+                                                                    Tensor,
                                                                     Tensor]:
         """Perform forward propagation/modelling.
 
@@ -109,66 +115,74 @@ class Elastic(torch.nn.Module):
         `mu`, `buoyancy`, and `grid_spacing` do not need to be provided again.
         See :func:`elastic` for a description of the inputs and outputs.
         """
-        return scalar(self.lamb, self.mu, self.buoyancy, self.grid_spacing, dt,
-                      source_amplitudes=source_amplitudes,
-                      source_locations=source_locations,
-                      receiver_locations=receiver_locations,
-                      accuracy=accuracy, pml_width=pml_width,
-                      pml_freq=pml_freq,
-                      max_vel=max_vel,
-                      survey_pad=survey_pad, 
-                      vy_0=vy_0,
-                      vx_0=vx_0,
-                      sigmayy_0=sigmayy_0,
-                      sigmaxy_0=sigmaxy_0,
-                      sigmaxx_0=sigmaxx_0,
-                      m_vyy_0=m_vyy_0,
-                      m_vyx_0=m_vyx_0,
-                      m_vxy_0=m_vxy_0,
-                      m_vxx_0=m_vxx_0,
-                      m_sigmayyy_0=m_sigmayyy_0,
-                      m_sigmaxyy_0=m_sigmaxyy_0,
-                      m_sigmaxyx_0=m_sigmaxyx_0,
-                      m_sigmaxxx_0=m_sigmaxxx_0,
-                      origin=origin, nt=nt,
-                      model_gradient_sampling_interval=
-                      model_gradient_sampling_interval)
+        return elastic(self.lamb, self.mu, self.buoyancy, self.grid_spacing,
+                       dt,
+                       source_amplitudes_y=source_amplitudes_y,
+                       source_amplitudes_x=source_amplitudes_x,
+                       source_locations_y=source_locations_y,
+                       source_locations_x=source_locations_x,
+                       receiver_locations_y=receiver_locations_y,
+                       receiver_locations_x=receiver_locations_x,
+                       accuracy=accuracy, pml_width=pml_width,
+                       pml_freq=pml_freq,
+                       max_vel=max_vel,
+                       survey_pad=survey_pad,
+                       vy_0=vy_0,
+                       vx_0=vx_0,
+                       sigmayy_0=sigmayy_0,
+                       sigmaxy_0=sigmaxy_0,
+                       sigmaxx_0=sigmaxx_0,
+                       m_vyy_0=m_vyy_0,
+                       m_vyx_0=m_vyx_0,
+                       m_vxy_0=m_vxy_0,
+                       m_vxx_0=m_vxx_0,
+                       m_sigmayyy_0=m_sigmayyy_0,
+                       m_sigmaxyy_0=m_sigmaxyy_0,
+                       m_sigmaxyx_0=m_sigmaxyx_0,
+                       m_sigmaxxx_0=m_sigmaxxx_0,
+                       origin=origin, nt=nt,
+                       model_gradient_sampling_interval=
+                       model_gradient_sampling_interval)
 
 
-def scalar(lamb: Tensor, mu: Tensor, buoyancy: Tensor,
-           grid_spacing: Union[int, float, List[float],
-                               Tensor],
-           dt: float,
-           source_amplitudes: Optional[Tensor] = None,
-           source_locations: Optional[Tensor] = None,
-           receiver_locations: Optional[Tensor] = None,
-           accuracy: int = 4, pml_width: Union[int, List[int]] = 20,
-           pml_freq: Optional[float] = None,
-           max_vel: Optional[float] = None,
-           survey_pad: Optional[Union[int, List[Optional[int]]]] = None,
-           vy_0: Optional[Tensor] = None,
-           vx_0: Optional[Tensor] = None,
-           sigmayy_0: Optional[Tensor] = None,
-           sigmaxy_0: Optional[Tensor] = None,
-           sigmaxx_0: Optional[Tensor] = None,
-           m_vyy_0: Optional[Tensor] = None,
-           m_vyx_0: Optional[Tensor] = None,
-           m_vxy_0: Optional[Tensor] = None,
-           m_vxx_0: Optional[Tensor] = None,
-           m_sigmayyy_0: Optional[Tensor] = None,
-           m_sigmaxyy_0: Optional[Tensor] = None,
-           m_sigmaxyx_0: Optional[Tensor] = None,
-           m_sigmaxxx_0: Optional[Tensor] = None,
-           origin: Optional[List[int]] = None, nt: Optional[int] = None,
-           model_gradient_sampling_interval: int = 1,
-           freq_taper_frac: float = 0.0,
-           time_pad_frac: float = 0.0) -> Tuple[Tensor, Tensor,
-                                                Tensor, Tensor,
-                                                Tensor, Tensor,
-                                                Tensor, Tensor,
-                                                Tensor, Tensor,
-                                                Tensor, Tensor,
-                                                Tensor, Tensor]:
+def elastic(lamb: Tensor, mu: Tensor, buoyancy: Tensor,
+            grid_spacing: Union[int, float, List[float],
+                                Tensor],
+            dt: float,
+            source_amplitudes_y: Optional[Tensor] = None,
+            source_amplitudes_x: Optional[Tensor] = None,
+            source_locations_y: Optional[Tensor] = None,
+            source_locations_x: Optional[Tensor] = None,
+            receiver_locations_y: Optional[Tensor] = None,
+            receiver_locations_x: Optional[Tensor] = None,
+            accuracy: int = 4, pml_width: Union[int, List[int]] = 20,
+            pml_freq: Optional[float] = None,
+            max_vel: Optional[float] = None,
+            survey_pad: Optional[Union[int, List[Optional[int]]]] = None,
+            vy_0: Optional[Tensor] = None,
+            vx_0: Optional[Tensor] = None,
+            sigmayy_0: Optional[Tensor] = None,
+            sigmaxy_0: Optional[Tensor] = None,
+            sigmaxx_0: Optional[Tensor] = None,
+            m_vyy_0: Optional[Tensor] = None,
+            m_vyx_0: Optional[Tensor] = None,
+            m_vxy_0: Optional[Tensor] = None,
+            m_vxx_0: Optional[Tensor] = None,
+            m_sigmayyy_0: Optional[Tensor] = None,
+            m_sigmaxyy_0: Optional[Tensor] = None,
+            m_sigmaxyx_0: Optional[Tensor] = None,
+            m_sigmaxxx_0: Optional[Tensor] = None,
+            origin: Optional[List[int]] = None, nt: Optional[int] = None,
+            model_gradient_sampling_interval: int = 1,
+            freq_taper_frac: float = 0.0,
+            time_pad_frac: float = 0.0) -> Tuple[Tensor, Tensor,
+                                                 Tensor, Tensor,
+                                                 Tensor, Tensor,
+                                                 Tensor, Tensor,
+                                                 Tensor, Tensor,
+                                                 Tensor, Tensor,
+                                                 Tensor, Tensor,
+                                                 Tensor]:
     """Elastic wave propagation (functional interface).
 
     This function performs forward modelling with the elastic wave equation.
@@ -179,12 +193,12 @@ def scalar(lamb: Tensor, mu: Tensor, buoyancy: Tensor,
 
     Args:
         lamb:
-            A 2D Tensor containing the first Lame parameter model, lambda.
+            A 2D Tensor containing the first Lamé parameter model, lambda.
             Unlike the module interface (:class:`Elastic`), in this
             functional interface a copy is not made of the model, so gradients
             will propagate back into the provided Tensor.
         mu:
-            A 2D Tensor containing the second Lame parameter model.
+            A 2D Tensor containing the second Lamé parameter model.
         buoyancy:
             A 2D Tensor containing the buoyancy (1/density) model.
         grid_spacing:
@@ -195,32 +209,50 @@ def scalar(lamb: Tensor, mu: Tensor, buoyancy: Tensor,
             A float specifying the time step interval of the input and
             output (internally a smaller interval may be used in
             propagation to obey the CFL condition for stability).
-        source_amplitudes:
-            A Tensor with dimensions [shot, source, time]. If two shots
-            are being propagated simultaneously, each containing three
-            sources of one hundred time samples, it would have shape
-            [2, 3, 100]. The time length will be the number of time steps
-            in the simulation. Optional. If provided, `source_locations` must
-            also be specified. If not provided, `nt` must be specified.
-        source_locations:
+        source_amplitudes_y:
+            A Tensor with dimensions [shot, source, time] containing time
+            samples of the source wavelets for sources oriented in the
+            first spatial dimension. If two shots are being propagated
+            simultaneously, each containing three sources, oriented in the
+            first spatial dimension, of one hundred time samples, it would
+            have shape [2, 3, 100]. The time length will be the number of
+            time steps in the simulation. Optional. If provided,
+            `source_locations_y` must also be specified. If not provided
+            (and `source_amplitudes_x` is also not specified), `nt`
+            must be specified.
+        source_amplitudes_x:
+            A Tensor containing source wavelet time samples for sources
+            oriented in the second spatial dimension. If `source_amplitudes_y`
+            is also specified, both must have the same length in the
+            shot and time dimensions.
+        source_locations_y:
             A Tensor with dimensions [shot, source, 2], containing the
             index in the two spatial dimensions of the cell that each
-            source is located in, relative to the origin of the model.
-            Optional. Must be provided if `source_amplitudes` is. It should
-            have torch.long (int64) datatype. The location of each source
-            must be unique within the same shot (you cannot have two
-            sources in the same shot that both have location [1, 2], for
-            example).
-        receiver_locations:
+            source oriented in the first spatial dimension is located in,
+            relative to the origin of the model. Optional. Must be provided
+            if `source_amplitudes_y` is. It should have torch.long (int64)
+            datatype. The location of each source oriented in the same
+            dimension must be unique within the same shot (you cannot have two
+            sources oriented in the first dimension in the same shot that
+            both have location [1, 2], for example).
+        source_locations_x:
+            A Tensor containing the locations of sources oriented in the
+            second spatial dimension.
+        receiver_locations_y:
             A Tensor with dimensions [shot, receiver, 2], containing
-            the coordinates of the cell containing each receiver. Optional.
+            the coordinates of the cell containing each receiver oriented
+            in the first spatial dimension. Optional.
             It should have torch.long (int64) datatype. If not provided,
-            the output `receiver_amplitudes` Tensor will be empty. If
+            the output `receiver_amplitudes_y` Tensor will be empty. If
             backpropagation will be performed, the location of each
-            receiver must be unique within the same shot.
+            receiver with the same orientation must be unique within the
+            same shot.
+        receiver_locations_x:
+            A Tensor containing the coordinates of the receivers oriented
+            in the second spatial dimension.
         accuracy:
             An int specifying the finite difference order of accuracy. Possible
-            values are 2, 4, 6, and 8, with larger numbers resulting in more
+            values are 2 and 4, with larger numbers resulting in more
             accurate results but greater computational cost. Optional, with
             a default of 4.
         pml_width:
@@ -236,10 +268,10 @@ def scalar(lamb: Tensor, mu: Tensor, buoyancy: Tensor,
             Larger values result in smaller reflections, with values of 10
             to 20 being typical. For a reflective or "free" surface, set the
             value for that edge to be zero. For example, if your model is
-            oriented so that surface that you want to be reflective is the
+            oriented so that the surface that you want to be reflective is the
             beginning of the second dimension, then you could specify
-            `pml_width=[20, 20, 0, 20]`. The wavespeed in the PML region
-            is obtained by replicating the values on the edge of the
+            `pml_width=[20, 20, 0, 20]`. The model values in the PML region
+            are obtained by replicating the values on the edge of the
             model. Optional, default 20.
         pml_freq:
             A float specifying the frequency that you wish to use when
@@ -377,41 +409,100 @@ def scalar(lamb: Tensor, mu: Tensor, buoyancy: Tensor,
             m_sigmayyy_0, m_sigmaxyy_0, m_sigmaxyx_0, m_sigmaxxx_0:
                 A Tensor containing the final value for the "memory variable"
                 for the stress, used in the PML.
-            receiver_amplitudes:
+            receiver_amplitudes_y:
                 A Tensor of dimensions [shot, receiver, time] containing
                 the receiver amplitudes recorded at the provided receiver
-                locations. If no receiver locations were specified then
-                this Tensor will be empty.
+                locations for receivers oriented in the first spatial
+                dimension. If no such receiver locations
+                were specified then this Tensor will be empty.
+            receiver_amplitudes_x:
+                A Tensor containing the receiver amplitudes for receivers
+                oriented in the second spatial dimension.
 
     """
-    empty_string_list: List[str] = []
-    (v, other_models, source_amplitudes, wavefields,
-     ay, ax, by, bx, sources_i, receivers_i,
+    if (source_locations_y is not None and
+            source_locations_y[..., 1].max() >= lamb.shape[1] - 1):
+        raise RuntimeError("With the provided model, the maximum y source "
+                           "location in the second dimension must be less "
+                           "than " + str(lamb.shape[1] - 1) + ".")
+    if (receiver_locations_y is not None and
+            receiver_locations_y[..., 1].max() >= lamb.shape[1] - 1):
+        raise RuntimeError("With the provided model, the maximum y "
+                           "receiver location in the second dimension "
+                           "must be less than " +
+                           str(lamb.shape[1] - 1) + ".")
+    if (source_locations_x is not None and
+            source_locations_x[..., 0].min() <= 0):
+        raise RuntimeError("The minimum x source "
+                           "location in the first dimension must be "
+                           "greater than 0.")
+    if (receiver_locations_x is not None and
+            receiver_locations_x[..., 0].min() <= 0):
+        raise RuntimeError("The minimum x receiver "
+                           "location in the first dimension must be "
+                           "greater than 0.")
+    if accuracy not in [2, 4]:
+        raise RuntimeError("The accuracy must be 2 or 4.")
+
+    (models, source_amplitudes, wavefields,
+     pml_profiles, sources_i, receivers_i,
      dy, dx, dt, nt, n_batch,
      step_ratio, model_gradient_sampling_interval,
      accuracy, pml_width_list) = \
-        setup_propagator(v, grid_spacing, dt, [], empty_string_list,
-                         [wavefield_0, wavefield_m1, psiy_m1, psix_m1,
-                          zetay_m1, zetax_m1],
-                         source_amplitudes,
-                         source_locations, receiver_locations,
+        setup_propagator([lamb, mu, buoyancy], 'elastic', grid_spacing, dt,
+                         [vy_0, vx_0, sigmayy_0, sigmaxy_0, sigmaxx_0,
+                          m_vyy_0, m_vyx_0, m_vxy_0, m_vxx_0,
+                          m_sigmayyy_0, m_sigmaxyy_0,
+                          m_sigmaxyx_0, m_sigmaxxx_0],
+                         [source_amplitudes_y, source_amplitudes_x],
+                         [source_locations_y, source_locations_x],
+                         [receiver_locations_y, receiver_locations_x],
                          accuracy, pml_width, pml_freq, max_vel,
                          survey_pad,
                          origin, nt, model_gradient_sampling_interval,
                          freq_taper_frac, time_pad_frac)
+    lamb, mu, buoyancy = models
+    source_amplitudes_y, source_amplitudes_x = source_amplitudes
+    (vy, vx, sigmayy, sigmaxy, sigmaxx,
+     m_vyy, m_vyx, m_vxy, m_vxx,
+     m_sigmayyy, m_sigmaxyy,
+     m_sigmaxyx, m_sigmaxxx) = wavefields
+    ay, ayh, ax, axh, by, byh, bx, bxh = pml_profiles
+    sources_y_i, sources_x_i = sources_i
+    receivers_y_i, receivers_x_i = receivers_i
+    if any([s <= (accuracy + 1) * 2 for s in lamb.shape]):
+        raise RuntimeError("The model must have at least " +
+                           str((accuracy + 1) * 2 + 1) +
+                           " elements in each dimension (including PML).")
 
-    wfc, wfp, psiy, psix, zetay, zetax, receiver_amplitudes = \
-        torch.ops.deepwave.scalar(
-            v, source_amplitudes, wavefields[0], wavefields[1],
-            wavefields[2], wavefields[3], wavefields[4], wavefields[5],
-            ay, ax, by, bx, sources_i, receivers_i, dy, dx, dt, nt,
+    (vy, vx, sigmayy, sigmaxy, sigmaxx,
+     m_vyy, m_vyx, m_vxy, m_vxx,
+     m_sigmayyy, m_sigmaxyy,
+     m_sigmaxyx, m_sigmaxxx, receiver_amplitudes_y,
+     receiver_amplitudes_x) = \
+        torch.ops.deepwave.elastic(
+            lamb, mu, buoyancy, source_amplitudes_y, source_amplitudes_x,
+            vy, vx, sigmayy, sigmaxy, sigmaxx,
+            m_vyy, m_vyx, m_vxy, m_vxx,
+            m_sigmayyy, m_sigmaxyy,
+            m_sigmaxyx, m_sigmaxxx,
+            ay, ayh, ax, axh, by, byh, bx, bxh,
+            sources_y_i, sources_x_i, receivers_y_i, receivers_x_i,
+            dy, dx, dt, nt,
             n_batch, step_ratio * model_gradient_sampling_interval,
             accuracy, pml_width_list[0], pml_width_list[1], pml_width_list[2],
             pml_width_list[3]
         )
 
-    receiver_amplitudes = downsample_and_movedim(receiver_amplitudes,
-                                                 step_ratio, freq_taper_frac,
-                                                 time_pad_frac)
+    receiver_amplitudes_y = downsample_and_movedim(receiver_amplitudes_y,
+                                                   step_ratio, freq_taper_frac,
+                                                   time_pad_frac)
+    receiver_amplitudes_x = downsample_and_movedim(receiver_amplitudes_x,
+                                                   step_ratio, freq_taper_frac,
+                                                   time_pad_frac)
 
-    return (wfc, wfp, psiy, psix, zetay, zetax, receiver_amplitudes)
+    return (vy, vx, sigmayy, sigmaxy, sigmaxx,
+            m_vyy, m_vyx, m_vxy, m_vxx,
+            m_sigmayyy, m_sigmaxyy,
+            m_sigmaxyx, m_sigmaxxx, receiver_amplitudes_y,
+            receiver_amplitudes_x)
